# Install the ensemble tools
from setuptools import setup, find_packages
import re
import sys
import os.path

# Use the Readme file as long description.
try:
    with open("README.md", "r") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = ""


def get_version():
    """
    read version string from enstools package without importing it

    Returns
    -------
    str:
            version string
    """
    with open("enstools/core/__init__.py") as f:
        for line in f:
            match = re.search(r'__version__\s*=\s*"([a-zA-Z0-9_.]+)"', line)
            if match is not None:
                return match.group(1)

def find_enstools_packages():
    """
    Find the packages inside the enstools folder.
    """

    return [f'enstools.{p}' for p in (find_packages(f'{os.path.dirname(__file__)}/enstools'))]


# only print the version and exit?
if len(sys.argv) == 2 and sys.argv[1] == "--get-version":
    print(get_version())
    exit()

# perform the actual install operation
setup(name="enstools",
      version=get_version(),
      author="Robert Redl et al.",
      author_email="robert.redl@lmu.de",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url="https://github.com/wavestoweather/enstools",
      packages=['enstools', *find_enstools_packages()],
      install_requires=[
          "appdirs",
          "numpy",
          "xarray",
          "cftime",
          "dask",
          "distributed",
          "cloudpickle",
          "colorama",
          "numba",
          "toolz",
          "pint",
          "cartopy",
          "matplotlib",
          "decorator",
          "multipledispatch",
          "cachey",
          "cffi",
          "pandas",
          "packaging",
          "h5netcdf",
          "h5py",
          "hdf5plugin",
          "numcodecs",
          "scikit-image",
          "scikit-learn",
          "scipy",
          "plotly",
          "bokeh",
          "statsmodels",
          "dataclasses",
      ],
      entry_points={
          'console_scripts': [
              'enstools-opendata=enstools.opendata.cli:main',
          ],
      },
      )



import numpy as np
import xarray


def mean_square_error(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description: Compute the mean square error.
    :param reference:
    :param target:
    :return:
    """
    non_temporal_dimensions = [d for d in reference.dims if d != "time"]
    return ((target - reference) ** 2).mean(dim=non_temporal_dimensions)


def root_mean_square_error(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description: A meaningful description of the metric with references if necessary would be nice.
    :param reference:
    :param target:
    :return:
    """
    return mean_square_error(reference, target) ** .5


# First we start defining few functions used in this file
def inter_quartile_range(array: xarray.DataArray) -> xarray.DataArray:
    """
    Returns the inter quartile range of a data array
    :param array:
    :return:
    """
    return array.quantile(0.75) - array.quantile(0.25)


def value_range(array: xarray.DataArray) -> xarray.DataArray:
    """
    Returns the value range of a data array
    :param array:
    :return:
    """
    return array.max() - array.min()


def normalized_root_mean_square_error(reference: xarray.DataArray, target: xarray.DataArray,
                                      method: str = "iqr") -> xarray.DataArray:
    """

    Description: Normalized RMSE with two normalization methods, absolute range and inter quartile range
    (less sensitive to outliers)
    :param reference:
    :param target:
    :param method: method used to normalize the root mean square error. Can be 'iqr' for inter quartile range or 'range'
    for the absolute value range.
    :return:
    """
    if method == "iqr":
        normalization_range = inter_quartile_range(reference)
    elif method == "range":
        normalization_range = value_range(reference)
    else:
        raise NotImplementedError

    if normalization_range != 0.:
        return root_mean_square_error(reference, target) / normalization_range
    else:
        return root_mean_square_error(reference, target)


def normalized_root_mean_square_error_index(reference: xarray.DataArray, target: xarray.DataArray,
                                            method="iqr") -> xarray.DataArray:
    """
    Description: Normalized Root Mean Square Error as index.
                For values that are really close to 0, we return the -log10 of the value, to have a better feeling on
                how close to the original value it is.

    Functions
    :param reference:
    :param target:
    :param method: normalization method used. Options are 'iqr' or 'range'
    :return:
    """
    nrmse = normalized_root_mean_square_error(reference=reference, target=target, method=method)
    return xarray.where(nrmse > 0, - np.log10(nrmse), np.inf)



def fix_attributes(data_array, suffix):
    """
    Fix a Data Array attributes adding a suffix to several of them.
    :param data_array:
    :param suffix:
    :return:
    """
    # Rewrite all the attributes:
    data_array.name = f"{data_array.name}_{suffix}"

    new_attributes = {
        "long_name": f"{suffix} of %s",
        "short_name": f"%s_{suffix}",
        "standard_name": f" %s_{suffix}",
    }
    for key, substitution in new_attributes.items():
        if key in data_array.attrs:
            data_array.attrs[key] = substitution % data_array.attrs[key]

    attributes_to_delete = ["code", "table"]

    for attr in attributes_to_delete:
        if attr in data_array.attrs:
            del (data_array.attrs[attr])
    return data_array



"""
Computation of basic ensemble statistics
"""
import xarray
from enstools.misc import get_ensemble_dim
from enstools.core.parallelisation import rechunk_arguments


@rechunk_arguments({"time": 1})
def ensemble_stat(dataset, stat=["mean", "min", "max", "std"], dim=None, ddof=1):
    """
    Compute ensemble mean and standard deviation from an input array or dataset.

    Parameters
    ----------
    dataset : xarray.DataArray or xarray.Dataset
            input data with ensemble dimension

    stat : list
            list auf standard statistics to compute. Implemented are "mean", "min", "max", "std".

    dim : str
            name of the ensemble dimension. If None, the name is determined automatically.

    ddof : int
            degrees of freedom in the calculation of the standard deviation.

    Returns
    -------
    tuple :
            (mean, min_, max_, std). Tuple with one item per item in argument stat.
    """
    # check the input data
    if not isinstance(dataset, xarray.DataArray) and not isinstance(dataset, xarray.Dataset):
        raise ValueError("the input to ensemble_stat has to be an xarray array or dataset!")

    # try to find the ensemble dimension
    if dim is None:
        dim = get_ensemble_dim(dataset)
        if dim is None:
            raise ValueError("ensemble_stat: unable to auto-detect the name of the ensemble dimension")

    # perform the calculation
    results = []
    for one_stat in stat:
        if one_stat == "mean":
            results.append(dataset.mean(dim=dim))
        if one_stat == "min":
            results.append(dataset.min(dim=dim))
        if one_stat == "max":
            results.append(dataset.max(dim=dim))
        if one_stat == "std":
            results.append(dataset.std(dim=dim, ddof=ddof))

    # return the result
    return tuple(results)



from .DWDRadar import getDWDRadar


def retrieve_radar(product=None, data_time=None, forecast_time=0, dest=None, file_format=None):
    """
    Downloads radar datasets from opendata server.
    Parameters
    ----------
        product: str
            The type of the radar data.

        data_time : list or int
            Time of the radar data. Multiple values are allowed.

        forecast_time : list or int
                time since the initialization of forecast. Multiple values are allowed.

        dest : str
                Destination folder for downloaded data. If the files are already available,
                they are not downloaded again.
        file_format: str
            The file format (eg. 'gz'). Only has to be specified when there are more than one available.

        Returns
        -------
        list :
                names of downloaded files.

    Returns
    -------
    list :
            names of downloaded files.
    """
    content = getDWDRadar()
    download_files = content.retrieve(product=product, data_time=data_time, forecast_time=forecast_time,
                                      dest=dest, file_format=file_format)
    return download_files



"""
A minial Interface to ecCodes based on CFFI
"""
import cffi
import numpy as np
import xarray
import struct
import threading
import platform
import logging
import os

# initialize the interface to the C-Library
ffi = cffi.FFI()

# type definition. No need to specify internals of the structs. just name them
ffi.cdef("typedef struct codes_handle codes_handle;")
ffi.cdef("typedef struct codes_context codes_context;")
ffi.cdef("typedef struct codes_keys_iterator codes_keys_iterator;")

# definition of the used functions
ffi.cdef("long codes_get_api_version (void);")
ffi.cdef("codes_handle* codes_handle_new_from_message(codes_context* c, const void*	data, size_t data_len);")
ffi.cdef("int codes_handle_delete(codes_handle* h);")
ffi.cdef("int codes_get_long(codes_handle* h, const char* key, long* value);")
ffi.cdef("int codes_get_double(codes_handle* h, const char* key, double* value);")
ffi.cdef("int codes_get_string(codes_handle* h, const char* key, char* mesg, size_t* length);")
ffi.cdef("int codes_get_size(codes_handle* h, const char* key, size_t* size);")
ffi.cdef("int codes_get_long_array(codes_handle* h, const char* key, long* vals, size_t* length);")
ffi.cdef("int codes_get_double_array(codes_handle* h, const char* key, double* vals, size_t* length);")
ffi.cdef("int grib_get_native_type(codes_handle* h, const char* name, int* type);")

# functions for key-iterators
ffi.cdef("codes_keys_iterator* codes_keys_iterator_new(codes_handle *h, unsigned long filter_flags, const char* name_space);")
ffi.cdef("int codes_keys_iterator_next (codes_keys_iterator *kiter);")
ffi.cdef("const char* codes_keys_iterator_get_name(codes_keys_iterator *kiter);")
ffi.cdef("int codes_keys_iterator_delete(codes_keys_iterator *kiter);")


def __find_lib_ld_library_path(name):
    """
    find a library on a linux system with the LD_LIBRARY_PATH, is defined.

    Parameters
    ----------
    name: str
            name of the library, e.g. libeccodes

    Returns
    -------
    str:
            absolute path of the library if found. if not found, the name is returned
    """
    LD_LIBRARY_PATH = os.getenv("LD_LIBRARY_PATH")
    if LD_LIBRARY_PATH is not None:
        components = LD_LIBRARY_PATH.split(":")
        for one_component in components:
            lib_path = os.path.join(one_component, name)
            if os.path.exists(lib_path + ".so"):
                return lib_path
    return name


# load the actual c-library
if platform.system() == "Linux":
    __libext = "so"
    __libname = __find_lib_ld_library_path("libeccodes")
elif platform.system() == "Darwin":
    __libext = "dylib"
    __libname = "libeccodes"
elif platform.system() == "Windows":
    __libext = "dll"
    __libname = "eccodes"
else:
    raise OSError("Unknown platform: %s" % platform.system())
try:
    _eccodes = ffi.dlopen("{name}.{ext}".format(name=__libname, ext=__libext))
except OSError:
    logging.warning("eccodes c-library not found, grib file support not available!")


# Constants for 'missing'
CODES_MISSING_DOUBLE = -1e+100
CODES_MISSING_LONG = 2147483647

# list of staggered variables in U-direction
# FIXME: this is COSMO-specific and related to issue #39
staggered_u = ["u", "aumfl_s"]  #, "u_10m", "umfl_s"]
staggered_v = ["v", "avmfl_s"]  #, "v_10m", "vmfl_s"]

# standard keys required by read_grib_file
standard_keys = ['bitsPerValue',
                 'cfName',
                 'cfVarName',
                 'dataDate',
                 'dataTime',
                 'discipline',
                 'editionNumber',
                 'gridDefinitionDescription',
                 'gridType',
                 'iDirectionIncrementInDegrees',
                 'indicatorOfParameter',
                 'jDirectionIncrementInDegrees',
                 'latitudeOfFirstGridPointInDegrees',
                 'latitudeOfLastGridPointInDegrees',
                 'latitudeOfSouthernPoleInDegrees',
                 'level',
                 'localActualNumberOfEnsembleNumber',
                 'longitudeOfFirstGridPointInDegrees',
                 'longitudeOfLastGridPointInDegrees',
                 'longitudeOfSouthernPoleInDegrees',
                 'missingValue',
                 'Ni',
                 'Nj',
                 'numberOfDataPoints',
                 'parameterCategory',
                 'parameterName',
                 'parameterNumber',
                 'parameterUnits',
                 'perturbationNumber',
                 'scaledValueOfFirstFixedSurface',
                 'scaledValueOfSecondFixedSurface',
                 'shortName',
                 'table2Version',
                 'typeOfLevel',
                 'validityDate',
                 'validityTime']

# allow only one read per time
read_msg_lock = threading.Lock()


# A representation of one grib message
class GribMessage():

    def __init__(self, file, offset=0, read_data=False):
        """
        create a message from the data buffer object

        Parameters
        ----------
        file : file-object
                a file object which points already to the beginning of the message

        offset : int
                position of the file where the message starts

        read_data : bool
                False: read only the header of the message.
        """
        # cache for all read operations of keys
        self.cache = {}
        self.has_data = read_data

        # read the content of the message
        self.buffer = _read_message_raw_data(file, offset, read_data=read_data)
        # was there a message?
        if self.buffer is None:
            self.handle = ffi.NULL
            return

        # decode the message
        with read_msg_lock:
            # read the message itself
            self.handle = _eccodes.codes_handle_new_from_message(ffi.NULL, ffi.from_buffer(self.buffer), len(self.buffer))

            # pre-read common keys and don't care for errors
            for one_key in standard_keys:
                try:
                    self.__getitem__(one_key, use_lock=False)
                except:
                    pass

            # pre-read the values also if we read the data in memory
            if read_data:
                self.__getitem__("values", use_lock=False)

        # was the reading successful?
        if self.handle == ffi.NULL:
            raise ValueError("unable to read grib message from buffer!")

    def __getitem__(self, item, use_lock=True):
        if item in self.cache:
            result = self.cache[item]
            if result is None:
                raise KeyError("key '%s' not found in grib message!" % item)
            return result
        else:
            try:
                # lock if we do not yet have a lock from a calling function
                if use_lock:
                    read_msg_lock.acquire()

                # read the key
                ckey = _cstr(item)
                nelements = self.__codes_get_size(ckey)
                if nelements > 1:
                    value = self.__codes_get_array(ckey, nelements)
                else:
                    value = self.__codes_get(ckey)
                self.cache[item] = value
            except ValueError:
                # store the error
                self.cache[item] = None
                # nothing found? Any error is interpreted as Key not found.
                raise KeyError("key '%s' not found in grib message!" % item)
            finally:
                # unlock if locked
                if use_lock:
                    read_msg_lock.release()
            return value

    def __contains__(self, item):
        # is the value already in cache?
        if item in self.cache:
            if self.cache[item] is None:
                return False
            return True
        else:
            # The value is not cached, try to read it from the grib message
            try:
                self.__getitem__(item)
            except KeyError:
                return False
            return True

    def keys(self):
        """
        returns all GRIB keys of this GRIB message

        Returns
        -------
        list :
                list of strings with the names of the keys
        """
        result = []
        with read_msg_lock:
            # 128 is the value of the C-constant GRIB_KEYS_ITERATOR_DUMP_ONLY and reduces the set of keys to those
            # really available
            kiter = _eccodes.codes_keys_iterator_new(self.handle, 128, ffi.NULL)
            while _eccodes.codes_keys_iterator_next(kiter) == 1:
                result.append(ffi.string(_eccodes.codes_keys_iterator_get_name(kiter)).decode("utf-8"))
        return result

    def is_valid(self):
        """
        returns true if the content of a message was readable
        """
        return self.buffer is not None

    def get_name(self, prefer_cf=True):
        """
        find a name for this variable.

        Parameters
        ----------
        prefer_cf : bool
                if True, the search order for the name is "cfName", "cfVarName", "shortName", otherwise it is
                "shortName", "cfName", "cfVarName".

        Returns
        -------
        string
                name of the variable.
        """
        if prefer_cf:
            name_keys = ["cfName", "cfVarName", "shortName"]
        else:
            name_keys = ["shortName", "cfName", "cfVarName"]
        for key in name_keys:
            result = self.__getitem__(key)
            if result != "unknown":
                break
        return result

    def get_dimension(self, dimensions=None, dimension_names=None):
        """
        get the shape of one message depending on the grid type

        Returns
        -------
        tuple
                (shape, dim-names)
        """
        if self["gridType"] == "rotated_ll":
            shape = (self["Nj"], self["Ni"])
            # the dimension names differ for staggered variables like u and v
            var_name = self["shortName"].lower()
            if var_name in staggered_u and self["typeOfLevel"] not in ["heightAboveSea", "isobaricInhPa"]:
                #breakpoint()
                dim_names = ["rlat", "srlon"]
            elif var_name in staggered_v and self["typeOfLevel"] not in ["heightAboveSea", "isobaricInhPa"]:
                dim_names = ["srlat", "rlon"]
            else:
                dim_names = ["rlat", "rlon"]
        elif self["gridType"] == "regular_ll":
            shape = (self["Nj"], self["Ni"])
            dim_names = ["lat", "lon"]
        elif self["gridType"] in ["sh", "reduced_gg", "unstructured_grid"]:
            shape = (self["numberOfDataPoints"],)
            dim_names = ["cell"]
        else:
            raise ValueError("don't know how to calculate the shape for grid type %s" % self["gridType"])

        # loop over all already used dims for comparison
        if dimensions is not None and dimension_names is not None:
            for one_var in dimensions.keys():
                if dimension_names[one_var] == dim_names and dimensions[one_var] != shape:
                    for id, dn in enumerate(dim_names):
                        dim_names[id] = "%s%d" % (dn, 2)
        return shape, dim_names

    def get_coordinates(self, dimension_names):
        """
        get the longitude and latitude coordinates for one message

        Returns
        -------
        tuple:
            ((lon-dim-names, lon-coord), (lat-dim-names), lat-coord)
        """
        # are coordinates available?
        if "longitudes" not in self or "latitudes" not in self:
            return None, None

        if self["gridType"] == "rotated_ll":
            lon = (dimension_names, np.array(self["longitudes"].reshape(self["Nj"], self["Ni"]), dtype=np.float32))
            lat = (dimension_names, np.array(self["latitudes"].reshape(self["Nj"], self["Ni"]), dtype=np.float32))
        elif self["gridType"] in ["sh", "reduced_gg", "unstructured_grid"]:
            lon = (dimension_names[0], np.array(self["longitudes"], dtype=np.float32))
            lat = (dimension_names[0], np.array(self["latitudes"], dtype=np.float32))
        elif self["gridType"] == "regular_ll":
            lon = (dimension_names[1], np.array(self["longitudes"].reshape(self["Nj"], self["Ni"])[0, :], dtype=np.float32))
            lat = (dimension_names[0], np.array(self["latitudes"].reshape(self["Nj"], self["Ni"])[:, 0], dtype=np.float32))
        else:
            lon = (dimension_names[1], np.array(self["longitudes"], dtype=np.float32))
            lat = (dimension_names[0], np.array(self["latitudes"], dtype=np.float32))
        return lon, lat

    def get_rotated_ll_info(self, dim_names):
        """
        get the rotated pole and the rotated lon/lat coordinates

        Parameters
        ----------
        dim_names : list
                names of the rlat and rlon dimensions

        Returns
        -------

        """
        if self["gridType"] != "rotated_ll":
            raise ValueError("The gridType '%s' has not rotated pole!" % self["gridType"])
        rotated_pole_name = "rotated_pole"
        if not dim_names[0].endswith("t"):
            rotated_pole_name += dim_names[0][-1]
        # create rotated pole description
        rotated_pole = xarray.DataArray(np.zeros(1, dtype=np.int8), dims=(rotated_pole_name,))
        rotated_pole.attrs["grid_mapping_name"] = "rotated_latitude_longitude"
        rotated_pole.attrs["grid_north_pole_latitude"] = self["latitudeOfSouthernPoleInDegrees"] * -1
        rotated_pole.attrs["grid_north_pole_longitude"] = self["longitudeOfSouthernPoleInDegrees"] - 180
        # create rotated coordinate arrays
        # perform calculations on large integers to avoid rounding errors
        factor = 10 ** 10
        first_lon = int(self["longitudeOfFirstGridPointInDegrees"] * factor)
        last_lon = int(self["longitudeOfLastGridPointInDegrees"] * factor)
        first_lat = int(self["latitudeOfFirstGridPointInDegrees"] * factor)
        last_lat = int(self["latitudeOfLastGridPointInDegrees"] * factor)
        if last_lon < first_lon and first_lon > 180 * factor:
            first_lon -= 360 * factor
        # using linspace instead of array and the stored increment to ensure the correct number of values.
        rlon_int = np.linspace(first_lon, last_lon, self["Ni"], dtype=np.int64)
        rlon = xarray.DataArray(np.asarray(rlon_int / factor, dtype=np.float32), dims=(dim_names[-1],))
        rlon.attrs["long_name"] = "longitude in rotated pole grid"
        rlon.attrs["units"] = "degrees"
        rlon.attrs["standard_name"] = "grid_longitude"
        rlat_int = np.linspace(first_lat, last_lat, self["Nj"], dtype=np.int64)
        rlat = xarray.DataArray(np.asarray(rlat_int / factor, dtype=np.float32), dims=(dim_names[-2],))
        rlat.attrs["long_name"] = "latitude in rotated pole grid"
        rlat.attrs["units"] = "degrees"
        rlat.attrs["standard_name"] = "grid_latitude"
        return rotated_pole_name, rotated_pole, rlat, rlon

    def get_level(self):
        """
        gets the center value of the level coordinate, or if available first and second layer

        """
        if self["typeOfLevel"] in ["generalVerticalLayer", "isobaricInhPa"]:
            return self["level"]
        if not "scaledValueOfFirstFixedSurface" in self or not "scaledValueOfSecondFixedSurface" in self:
            return self["level"]
        first_surface = self["scaledValueOfFirstFixedSurface"]
        second_surface = self["scaledValueOfSecondFixedSurface"]
        first_missing = first_surface == CODES_MISSING_LONG or first_surface == CODES_MISSING_DOUBLE
        second_missing = second_surface == CODES_MISSING_LONG or second_surface == CODES_MISSING_DOUBLE

        if first_missing and not second_missing:
            return second_surface
        elif not first_missing and second_missing:
            return first_surface
        elif first_missing and second_missing:
            return self["level"]
        else:
            return first_surface, second_surface

    def get_values(self, shape=None, dtype=None, missing=None):
        """
        read the encoded values from the message

        Parameters
        ----------
        dtype : np.dtype
                values are returned in an array of the specified type

        missing : float
                value used within the grib message the mark missing values. The returned array will contain NaN at this
                locations.

        Returns
        -------
        np.ndarray
        """
        # do we have data in this message?
        if not self.has_data:
            raise ValueError("this message was created from the header only. No data is available!")
        values = self["values"]
        if shape is not None:
            values = values.reshape(shape)
        if dtype is not None and dtype != np.float64:
            values = np.array(values, dtype=dtype)
        # replace fill values with nan
        values = np.where(values == missing, np.nan, values)
        return values

    def __grib_get_native_type(self, key):
        """
        Get the native type of a specific grib key
        """
        itype = ffi.new("int[1]")
        err = _eccodes.grib_get_native_type(self.handle, key, itype)
        if err != 0:
            raise ValueError("unable to get type of key '%s'" % ffi.string(key))
        if itype[0] == 1:
            return int
        elif itype[0] == 2:
            return float
        else:
            return str

    def __codes_get_size(self, key):
        """
        get the number of elements for a given key

        Parameters
        ----------
        key : cstr
                name of the key

        Returns
        -------
        int :
                number of elements
        """
        size = ffi.new("size_t[1]")
        err = _eccodes.codes_get_size(self.handle, key, size)
        if err != 0:
            raise ValueError("unable to get number of elements for key '%s'" % ffi.string(key))
        return size[0]

    def __codes_get(self, key):
        """
        get the value of a non-array key

        Parameters
        ----------
        key : cstr
                name of the key

        Returns
        -------
        int or float or str
        """
        key_type = self.__grib_get_native_type(key)
        if key_type == int:
            value_ptr = ffi.new("long[1]")
            err = _eccodes.codes_get_long(self.handle, key, value_ptr)
            value = value_ptr[0]
        elif key_type == float:
            value_ptr = ffi.new("double[1]")
            err = _eccodes.codes_get_double(self.handle, key, value_ptr)
            value = value_ptr[0]
        else:
            value_buffer = np.zeros(1024, dtype=np.uint8)
            value_buffer_length = ffi.new("size_t[1]", init=[1024])
            err = _eccodes.codes_get_string(self.handle, key, ffi.from_buffer(value_buffer), value_buffer_length)
            if value_buffer_length[0] == 1024:
                value_buffer_length[0] = np.where(value_buffer == 0)[0][0]
            value = value_buffer[:value_buffer_length[0]-1].tobytes().decode("utf-8")
        if err != 0:
            raise ValueError("unable to get value for key '%s'" % ffi.string(key))
        return value

    def __codes_get_array(self, key, nelements):
        """
        Get a values for a key with multiple values

        Parameters
        ----------
        key : cstr
                name of the key

        nelements : int
                size the array to retrieve

        Returns
        -------
        np.ndarray
        """
        key_type = self.__grib_get_native_type(key)
        length = ffi.new("size_t[1]")
        length[0] = nelements
        if key_type == int:
            values = np.empty(nelements, dtype=np.int64)
            err = _eccodes.codes_get_long_array(self.handle, key, ffi.cast("long*", ffi.from_buffer(values)), length)
        elif key_type == float:
            values = np.empty(nelements, dtype=np.float64)
            err = _eccodes.codes_get_double_array(self.handle, key, ffi.cast("double*", ffi.from_buffer(values)), length)
        else:
            raise ValueError("string arrays are not yet supported!")
        if err != 0:
            raise ValueError("unable to get value for key '%s'" % ffi.string(key))
        return values

    def __del__(self):
        """
        free up the memory
        """
        with read_msg_lock:
            if self.handle != ffi.NULL:
                err = _eccodes.codes_handle_delete(self.handle)
                self.handle = ffi.NULL
                if err != 0:
                    raise ValueError("unable to free memory of grib message!")


def _cstr(pstr):
    """
    convert a python string object into a c string object (copy).

    Parameters
    ----------
    pstr : str
            python string

    Returns
    -------
    const char*
    """
    buffer = np.frombuffer(pstr.encode("utf8") + b"\x00", dtype=np.uint8)
    result = ffi.from_buffer(buffer)
    return result


def _read_message_raw_data(infile, offset, read_data=False):
    """
    Read the header of a grib message and return an byte array with the length of the full message, but without
    the actual data

    Parameters
    ----------
    infile

    Returns
    -------

    """
    # find the start word GRIB. Allow up to 1k junk in front of the actual message
    infile.seek(offset)
    start = infile.read(1024)
    istart = start.find(b"GRIB")
    if istart == -1:
        return None
    offset += istart

    # find at first the grib edition to account for different formats
    infile.seek(offset + 7)
    edition = struct.unpack(">B", infile.read(1))[0]

    # get the length of the total message
    if edition == 1:
        # read the first section
        infile.seek(offset)
        section0 = infile.read(8)
        length_total = struct.unpack(">I", b'\x00' + section0[4:7])[0]

        # check if the length is correct, the message is supposed to end with 7777
        # this is a workaround, apparently, the length of grib1 messages is sometimes wrong.
        infile.seek(offset + length_total - 4)
        section5 = infile.read(4)
        if section5 != b"7777":
            # the maximal length of a grib1 message is 16MB. Read this amount of data and search for the end
            infile.seek(offset)
            maxdata = infile.read(16777216)
            endpos = maxdata.find(b"7777")
            if endpos == -1:
                return None
            else:
                length_total = endpos + 4
                read_data = True
        infile.seek(offset + 8)

        # create an numpy array with the total size of the message
        bytes = np.zeros(length_total, dtype=np.uint8)
        bytes[0:8] = np.frombuffer(section0, dtype=np.uint8)
        pos = 8

        # read the complete message?
        if read_data:
            infile.readinto(memoryview(bytes[8:]))
            return bytes

        # read the first sections, but not the data
        for sec in range(1, 5):
            # read the length of the section
            infile.readinto(memoryview(bytes[pos:pos+3]))
            length_sec = struct.unpack(">I", b'\x00' + bytes[pos:pos+3].tobytes())[0]

            # do not read if this is the final data section
            if pos + length_sec + 4 >= length_total:
                # read the first bytes only
                infile.readinto(memoryview(bytes[pos+3:pos+11]))
                infile.seek(offset + length_total - 5)
                infile.readinto(memoryview(bytes[-5:]))
                break
            else:
                # read data of this section
                infile.readinto(memoryview(bytes[pos+3:pos+length_sec]))
                pos = pos + length_sec
    else:
        # read first section
        infile.seek(offset)
        section0 = infile.read(16)
        length_total = struct.unpack(">Q", section0[8:16])[0]

        # create an numpy array with the total size of the message
        bytes = np.zeros(length_total, dtype=np.uint8)
        bytes[0:16] = np.frombuffer(section0, dtype=np.uint8)
        pos = 16

        # read the complete message?
        if read_data:
            infile.readinto(memoryview(bytes[16:]))
            return bytes

        # read the first sections, but not the data.
        # For standard binary data, we don't read the data section. For other formats we do to avoid decoding errors.
        # TODO: replace any complex data representation with standard binary data to avoid reading!
        data_representation = 0
        while True:
            # read the length of the section and the section number
            infile.readinto(memoryview(bytes[pos:pos+5]))
            length_sec = struct.unpack(">I", bytes[pos:pos+4].tobytes())[0]
            section = bytes[pos+4]

            # do not read completely if this is the final data section
            if pos + length_sec + 4 >= length_total:
                # read the first bytes only (max 1k). For any complex packing, read the complete data section.
                if data_representation != 0:
                    infile.readinto(memoryview(bytes[pos+5:pos+length_sec]))
                # read the final 7777
                infile.seek(offset + length_total - 4)
                infile.readinto(memoryview(bytes[-4:]))
                break
            else:
                # read data of this section
                infile.readinto(memoryview(bytes[pos+5:pos+length_sec]))
                # if this is section 5, get the data representation type
                if section == 5:
                    data_representation = struct.unpack(">H", bytes[pos+9:pos+11].tobytes())[0]
                pos = pos + length_sec

    return bytes



"""
some service functions for interaction with OS environments
"""
import six
import os
import re
import socket
from contextlib import closing
from threading import Thread
from subprocess import Popen, PIPE, STDOUT
import logging
import appdirs

if six.PY2:
    from commands import getstatusoutput
else:
    from subprocess import getstatusoutput


def getenv(name):
    """
    get the value of an environment variable.

    Parameters
    ----------
    name : str
            name of the environment variable

    Raises
    ------
    KeyError :
            raises a KeyError if the Variable was not found in the environment

    Returns
    -------
    the value
    """
    value = os.getenv(name)
    if value is None:
        raise KeyError("Environment Variable not found: %s" % name)
    return value


def get_ip_address(interface=None):
    """
    Get an IP-Address to be used for dask communication. If an interface is given, the IP of this interface is used.
    Otherwise the fastest interface will be used.

    Parameters
    ----------
    interface : str
            Name of the interface to be used.

    Returns
    -------
    str:
            IP-Address to be used for communication.
    """
    # use socket module to get the IP associated with the hostname
    ip_hostname = socket.gethostbyname(socket.gethostname())

    # get a list of all interfaces if none was specified. The list is stored in a dictionary which will transport
    # all information about one interface
    interfaces = {}
    if interface is None:
        sts, out = getstatusoutput("ip link show up")
        if sts != 0:
            raise OSError("unable to get list of interfaces: %s" % out)
        # extract a list of all interfaces from the output
        interface_lines = re.split(r"^\d+: ", out, flags=re.MULTILINE)
        for line in interface_lines[1:]:
            iface = line.split(":", 1)[0]
            interfaces[iface] = {}
    else:
        interfaces[interface] = {}
    logging.debug("get_ip_address: list of interfaces: %s" % ", ".join(interfaces.keys()))

    # get the IPs for all interfaces
    for one_interface in interfaces.keys():
        sts, out = getstatusoutput("ip addr show %s" % one_interface)
        if sts != 0:
            raise OSError("unable to the ip address for interface %s" % one_interface)
        addr_match = re.search(r"inet (\d+.\d+.\d+.\d+)", out)
        if addr_match is None:
            logging.debug("get_ip_address: no IP address found for interface %s" % one_interface)
            interfaces[one_interface]["ip"] = None
        else:
            interfaces[one_interface]["ip"] = addr_match.group(1)
        logging.debug("get_ip_address: IP for interface %s: %s" % (one_interface, interfaces[one_interface]["ip"]))

    # find the fastest interface
    fastest_speed = -9999
    fastest_interfaces = None
    logging.debug("get_ip_address: speed of interfaces:")
    for one_interface in interfaces.keys():
        try:
            with open("/sys/class/net/%s/speed" % one_interface, "r") as f:
                interfaces[one_interface]["speed"] = int(f.read())
        except IOError:
            interfaces[one_interface]["speed"] = 0
        if interfaces[one_interface]["speed"] == fastest_speed:
            fastest_interfaces.append(one_interface)
        if interfaces[one_interface]["speed"] > fastest_speed:
            fastest_speed = interfaces[one_interface]["speed"]
            fastest_interfaces = [one_interface]
        logging.debug("get_ip_address: interfaces %s: %d" % (one_interface, interfaces[one_interface]["speed"]))

    # if there are more interfaces with the same speed, check if one has the IP associated with the hostname
    fastest_interface = fastest_interfaces[0]
    if len(fastest_interfaces) > 1:
        logging.debug("get_ip_address: found %d interfaces with same speed:" % len(fastest_interfaces))
        for iface in fastest_interfaces:
            logging.debug("get_ip_address: %s, IP: %s" % (iface, interfaces[iface]["ip"]))
            if interfaces[iface]["ip"] is not None and ip_hostname == interfaces[iface]["ip"]:
                logging.debug("get_ip_address: %s is associated with the hostname, selecting this interface!" % (iface))
                fastest_interface = iface
                break

    # log selection
    logging.debug("get_ip_address: selected interface:  %s" % fastest_interface)
    logging.debug("get_ip_address: selected IP-address: %s" % interfaces[fastest_interface]["ip"])
    return interfaces[fastest_interface]["ip"]


def get_first_free_port(host, start_port):
    """
    find a free port on the given host on which the scheduler can be started

    Parameters
    ----------
    host
    start_port

    Returns
    -------
    int :
            the first free port above the start_port
    """
    while True:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            try:
                s.bind((host, start_port))
            except:
                start_port +=1
                continue
            return s.getsockname()[1]


def which(cmd):
    """
    use the unix command which to find an executable

    Parameters
    ----------
    cmd : str
            name of the executable to search

    Returns
    -------
    str:
            absolute path of the executable
    """
    sts, out = getstatusoutput("which %s" % cmd)
    if sts != 0:
        raise IOError("executable not found: %s" % cmd)
    return out


def get_cache_dir():
    """
    create a platform depended cache directory.

    Returns
    -------
    str:
            path to the cache directory.
    """
    cache_dir = appdirs.user_cache_dir(appname="enstools")
    if not os.path.exists(cache_dir):
        logging.debug(f"Created cache directory: {cache_dir}")
        os.makedirs(cache_dir)
    return cache_dir


class ProcessObserver(Thread):
    """
    run a process and store its stdout
    """
    def __init__(self, args):
        """
        create an observer for a given command and start the process

        Parameters
        ----------
        cmd : str
                command to start and to follow
        """
        super(ProcessObserver, self).__init__()
        self.args = args
        self.on_exit_args = None
        self.p = None
        self.oep = None
        self.setDaemon(True)
        self.start()

    def run_on_exit(self, args):
        """
        an additional command to run after the main process ended (e.g. for cleanup)

        Parameters
        ----------
        args : list
                arguments of the program as expected by Popen
        """
        self.on_exit_args = args

    def run(self):
        """
        start the process and store it's output
        """
        # start the process
        self.p = Popen(self.args, stdin=PIPE, stdout=PIPE, stderr=STDOUT)

        # observe the output of the process as long as it is running
        while self.p.poll() is None:
            line = self.p.stdout.readline()
            if len(line) == 0:
                continue
            line = line.rstrip()
            # INFO output from dask processes should only be shown in debug mode.
            if b"- INFO -" in line:
                logging.debug(line)
            else:
                print(line)
        logging.debug("%s exited with exit code %d" % (self.args[0], self.p.returncode))
        if self.p.returncode != 0:
            logging.debug(self.p.stdout.read())

        # if there is an command specified to be executed at the end of the thread, then do it now
        if self.on_exit_args is not None:
            self.oep = Popen(self.on_exit_args)
            self.oep.wait()

    def poll(self):
        """
        the the process state. If an on exit process is defined, poll that as well.
        """
        p_poll = self.p.poll()
        if self.oep is not None:
            oep_poll = self.oep.poll()
        else:
            oep_poll = 0
        if p_poll is None:
            return None
        else:
            if oep_poll is None:
                return None
            else:
                return p_poll + oep_poll

    def terminate(self):
        return self.p.terminate()

    @property
    def returncode(self):
        return self.p.returncode




import xarray
from scipy.stats import ks_2samp
import numpy as np
from typing import List
from .fix_attributes import fix_attributes


def kolmogorov_smirnov_multicell(reference: xarray.DataArray, target: xarray.DataArray, number_of_cells: int = 100,
                                 ensemble_dimension="ens") \
        -> xarray.DataArray:
    """
    Description:     Compute the Kolmogorov-Smirnov statistic in a random sample of cells.

    Functions
    :param reference:
    :param target:
    :param number_of_cells
    :param ensemble_dimension name of the ensemble dimension, usually 'ens' or 'member'
    :return:
    """
    # Even if we are selecting only few cells, it might be convenient to just load the datasets to speedup things:
    reference.load()
    target.load()
    # Create a data array from the reference.
    result = reference.copy(deep=True)
    non_temporal_dimensions = [d for d in result.dims if d != "time"]
    selector = {dim: reference[dim][0] for dim in non_temporal_dimensions}
    result = result.sel(selector)

    # Randomly select a certain number of cells
    random_cells = random_select_cells(reference, ensemble_dimension=ensemble_dimension, size=number_of_cells)
    if "time" in reference.dims:
        times = reference["time"]
        statistics_time_series = []
        for t in times:
            statistics = [
                ks_2samp(target.sel(time=t).sel(sl).values.ravel(), reference.sel(time=t).sel(sl).values.ravel(),
                         alternative='two-sided', mode='auto').statistic for sl in random_cells]
            statistics_time_series.append(np.mean(statistics))
        result.values = statistics_time_series
    else:
        statistics = [ks_2samp(target.sel(sl).values.ravel(), reference.sel(sl).values.ravel(),
                               alternative='two-sided', mode='auto').statistic for sl in random_cells]
        result.values = np.mean(statistics)

    result = fix_attributes(result, "ks_multicell")
    return result


def random_select_cell(data_array: xarray.DataArray, ensemble_dimension: str = "ens") -> dict:
    """
    Uses numpy to randomly select a cell.
    Since xarray.DataArray are used, we identify each cell by a dictionary containing a value for each dimension.

    #TODO: Must check if this selection mode is actually slower than using indices.

    :param data_array:
    :param ensemble_dimension:
    :return:
    """
    dimensions_to_ignore = ["time", ensemble_dimension]

    selection = {dim: np.random.choice(data_array[dim]) for dim in data_array.dims if dim not in dimensions_to_ignore}

    return selection


def random_select_cells(data_array: xarray.DataArray, ensemble_dimension: str = "ens", size: int = 1) -> List[dict]:
    """
    Randomly select a certain number of cells
    :param data_array:
    :param ensemble_dimension:
    :param size:
    :return:
    """
    return [random_select_cell(data_array, ensemble_dimension=ensemble_dimension) for _ in range(size)]



from enum import Enum, auto
from abc import ABCMeta, abstractmethod

from enstools.misc import generate_coordinates, spherical2cartesian
from enstools.interpolation import nearest_neighbour

import numpy as np
from scipy.spatial import cKDTree
from math import sqrt, floor
from collections import OrderedDict

import cartopy
import matplotlib.pyplot as plt



class Backend(Enum):
    """
    Enum class that defining a library for plotting.
    """
    PLOTLY = auto()
    BOKEH  = auto()
    
class Stream(Enum):
    """
    Enum class that defining a type of streamlines.
    """
    LINE   = auto()
    VECTOR = auto()
    
class Function(Enum):
    """
    Enum class that defining a Plotly function used for a plotting.
    """
    TRISURF = auto()
    MESH3D  = auto()
    

class InteractiveBackend(metaclass=ABCMeta):
    """
    Abstract backend class that will be implemented for different plotting backends. Functions that are used for all
    backends can be implemented directly here.
    """
        
    @classmethod
    def icosahedral_to_regular_coords(cls, data_icosahedral, clon, clat, cell_area=None):
        """
        Converts icosahedral coordinates to regular-lat-lon coordinates.

        Parameters
        ----------
        data_icosahedral : xarray
            Values of a variable in icosahedral coordinates.

        clon: xarray
            Longitude coordinates

        clat: xarray
            Latitude coordinates

        cell_area: xarray
            Array containing the area of all cells.

        Returns
        -------
        xarray
            Values of a variable in regular-lat-lon coordinates.
        """
        
        # estimate regular cell size based on grid cell area
        if cell_area is not None:
            cell_size = sqrt(np.mean(cell_area))
        else:
            # calculate mean distance for a number of selected points
            cartesian_coords = spherical2cartesian(clon, clat)
            kdtree = cKDTree(cartesian_coords)
            grid_point_distance, _ = kdtree.query(cartesian_coords[[0, len(cartesian_coords) // 2, len(cartesian_coords) - 1]], k=2)
            # assumption: equilateral triangle
            cell_size = np.sqrt(1.299 * grid_point_distance[:, 1].mean() ** 2)

        # calculate resolution based on estimated number of cells
        perimeter = 2 * np.pi * 6371229.0
        ncell = perimeter / cell_size
        resolution = 360.0 / ncell
        
        # define borders of the region
        lon0 = np.degrees(np.min(clon))
        lon1 = np.degrees(np.max(clon))
        lat0 = np.degrees(np.min(clat))
        lat1 = np.degrees(np.max(clat))
        
        # convert to regular-lon-lat coordinates 
        lon, lat = generate_coordinates(resolution, lon_range=[lon0, lon1], lat_range=[lat0, lat1])
        interpol = nearest_neighbour(clon, clat, lon, lat, src_grid="unstructured", dst_grid="regular")
        data_regular = interpol(data_icosahedral)
        return data_regular
    
    @classmethod
    def coords_name(cls, data):
        """
        Defines exact names of regular-lat-lon coordinates of a xarray object.

        Parameters
        ----------
        data : xarray
            Array contains values of variable and regular-lat-lon coordinates.

        Returns
        -------
        lon, lat: str
            Name of the longitude or latitude coordinate in the xarray.
        """
        
        # names for coordinates
        __lon_names = ["lon", "lons", "longitude", "longitudes", "clon", "rlon"]
        __lat_names = ["lat", "lats", "latitude", "latitudes", "clat", "rlat"]
        
        for dim_name in data.dims:
            if dim_name in __lon_names:
                lon = dim_name
            elif dim_name in __lat_names:
                lat = dim_name
                
        return lon, lat
    
    @classmethod
    def get_aspect_ratio(cls, lon0, lon1, lat0, lat1):
        """
        Defines an aspectratio of a figure based on a size of a region.

        Parameters
        ----------            
        lon0, lon1, lat0, lat1 : float
            Values of longitudes or latitudes of a defined region borders. 

        Returns
        -------
        float
            Aspectratio for a figure.
        """    
        
        lon_range = lon1 - lon0
        lat_range = lat1 - lat0
        aspectratio = lon_range / lat_range
        
        return aspectratio
        
    @classmethod
    def make_path_map(cls, feature, map_resolution):
        """
        Defines paths of coastlines or country borders.

        Parameters
        ----------
        feature : {'coastlines', 'borders'}
            Argument 'coastlines' for paths of coastlines. Argument 'borders' for paths of country borders.
            
        map_resolution : {'110m', '50m', '10m'}
            Definition of a map resolution.

        Returns
        -------
        xs, ys : list of lists
            Lists of lists contain separated X- or Y-coordinates of paths.
        """

        paths = []       
        # create paths
        if feature == 'coastlines':
            for geom in cartopy.feature.COASTLINE.with_scale(map_resolution).geometries():
                if geom.geom_type == 'LineString':
                    paths.append(list(geom.coords))
                else:
                    for g in geom.geoms:
                        paths.append(list(g.coords))
                    
        elif feature == 'borders':
            for geom in cartopy.feature.BORDERS.with_scale(map_resolution).geometries():
                if geom.geom_type == 'LineString':
                    paths.append(list(geom.coords))
                else:
                    for g in geom.geoms:
                        paths.append(list(g.coords))
                    
        else:
            raise ValueError(f"unsupported feature: {feature}")
               
        xs = []
        ys = []
        # separate X- and Y-coordinates of paths
        for i in range(len(paths)):
            coords = np.array([(vertex[0], vertex[1]) for vertex in paths[i]])
            lons = coords.T[0].tolist()
            lats = coords.T[1].tolist()
            xs.append(lons)
            ys.append(lats)

        return xs, ys
    
    @classmethod
    def get_contour_data(cls, x, y, z, colormap, filled_contours=True):
        """
        Gets parameters of contours created using Matplotlib.pyptlot. 

        Parameters
        ----------
        x, y : Numpy array
            Arrays contain X- or Y-coordinates of grid cells.
                      
        z : Numpy array
            2-dimensional array contains values of a variable for every grid cell.
                       
        colormap : str
            Name of a colormap.
            
        filled_contours : bool, optional, default=True
            Definition of contours type. Argument equals to True for filled contours and 
            equals to False for unfilled contours.
                      
        Returns
        -------
        xs, ys : list of lists
            Lists of lists contain separated X- or Y-coordinates of contours paths.
                
        colors : list of str
            List contains colors of contours in HEX format.
                
        values : list of str
            List contains values of contours.      
        """
        
        xs = []
        ys = []
        colors = []
        values = []
        
        # create contours using Matplotlib.pyptlot
        if filled_contours: 
            cs = plt.contourf(x, y, z, cmap=colormap)
        else:
            cs = plt.contour(x, y, z, cmap=colormap)
            
        # get colors and values of contours
        isolevelid = 0
        for isolevel in cs.collections:
            if filled_contours: 
                isocol = isolevel.get_facecolors()[0]
            else:
                isocol = isolevel.get_color()[0]  
            thecol = 3 * [None]
            theiso = str(cs.get_array()[isolevelid])
            isolevelid += 1

            for i in range(3):
                thecol[i] = int(255 * isocol[i])
            thecol = '#%02x%02x%02x' % (thecol[0], thecol[1], thecol[2])

            for path in isolevel.get_paths():
                colors.append(thecol)
                values.append(theiso)
                
        # get paths of contours        
        for i in range(len(cs.allsegs)):
            for j in range(len(cs.allsegs[i])):
                xs.append(cs.allsegs[i][j].T[0].tolist())
                ys.append(cs.allsegs[i][j].T[1].tolist())
                
        # close unclosed paths for a filled contours case   
        if filled_contours:
            # check which contours are closed
            closed_index = []
            for i in range(len(xs)):
                if xs[i][0] == xs[i][-1]:
                    closed_index.append(i)
            closed_index_np = np.asarray(closed_index)
            
            # rewrite contours to have only closed parts of contours
            xs_unclosed = xs
            ys_unclosed = ys
            xs = []
            ys = []
            for i in range(len(xs_unclosed)):
                # make no modifications for closed contours
                if i in closed_index_np:
                    xs.append(xs_unclosed[i])
                    ys.append(ys_unclosed[i])
                # cut off unclosed part of unclosed contours
                else:
                    index = []
                    for j in range(len(xs_unclosed[i])):
                        for l in range(j+1, len(xs_unclosed[i])):
                            if xs_unclosed[i][j] == xs_unclosed[i][l] and ys_unclosed[i][j] == ys_unclosed[i][l]:
                                index.append(j)
                                index.append(l-1)
                    index_np = np.asarray(index)
                    xs_closed = xs_unclosed[i][index_np[0]:index_np[1]]
                    ys_closed = ys_unclosed[i][index_np[0]:index_np[1]]
                    xs.append(xs_closed)
                    ys.append(ys_closed)
            
        plt.close()

        return xs, ys, colors, values
    
    @classmethod
    def generate_colorscale(cls, colors, values):
        """
        Generates ordered sets of unique colors and unique values.
    
        Parameters
        ----------
        colors : list of str
            Set of colors. List of HEX or RGB color strings.
                 
        values : list of str
            Set of values.

        Returns
        -------
        colorset : list of tuples
            Set of ordered unique colors. List containing tuples mapping a normalized color values. 
                 
        valueset : Numpy array
            Set of ordered unique float values.
                 
        tick_interval : float
            Value of intervals between ticks in a colorbar.
        """
        
        # select unique colors and generate a normalized colorset
        colors_unique = [list(OrderedDict.fromkeys(colors).keys())]
        
        colorset = []
        divisions = 1. / len(colors_unique) 
        c_index = 0.
        for cset in colors_unique:
            nsubs = len(cset)
            sub_divisions = divisions / nsubs
            for subcset in cset:
                colorset.append((c_index,subcset))
                colorset.append((c_index + sub_divisions - .001,subcset))
                c_index = c_index + sub_divisions
        colorset[-1] = (1, colorset[-1][1])
        
        # select unique values and get a tick interval
        valueset = np.unique(np.asarray(values, dtype=np.float32))
        tick_interval = valueset[1] - valueset[0]

        return colorset, valueset, tick_interval
    
    @classmethod
    def get_streamline_data(cls, x, y, u, v, density):
        """
        Gets paths of streamlines created using Matplotlib.pyplot.

        Parameters
        ----------
        x, y : xarray
            Arrays contain X- or Y- coordinates of grid cells.
                  
        u, v : xarray
            2-dimensional arrays contain values of horizontal or vertical components of a variable.
            
        density : float
            Controls the closeness of vectors. When density = 1, the domain is divided into a 30x30 grid.
              
        Returns
        -------
        Numpy array
            3-dimensional array contains X- and Y-coordinates of streamlines.
        """
        
        # interpolate over 2D grid
        lons = np.linspace(x.min(), x.max(), x.size)
        lats = np.linspace(y.min(), y.max(), y.size)
        
        # get paths of streamlines created by Matplotlib.pyplot
        stream = plt.streamplot(lons, lats, u, v, density=density)
        streamlines_path = np.asarray(stream.lines.get_segments()) 
    
        plt.close()
        
        return streamlines_path
        
    @classmethod
    def reduce_vector_data(cls, x, y, u, v, density):
        """
        Gets parameters for a reduced number of vectors.

        Parameters
        ----------
        x, y : xarray.DataArray
            Arrays contain X- or Y-coordinates of grid cells.

        u, v : xarray.DataArray
            2-dimensional arrays contain values of horizontal or vertical components of a variable.
            
        density : float
            Controls the closeness of vectors. When density = 1, the domain is divided into a 30x30 grid.

        Returns
        -------
        lons, lats : Numpy array
            2-dimensional array contain X- or Y-coordinates of a reduced number of grid cells.

        us, vs : Numpy array
            2-dimensional arrays contain reduced values of horizontal or vertical components of a variable.
        """
        
        # reduce a number of vectors
        nlon = int(density*30)
        nlat = int(density*30)
        if nlon > x.size and nlat > y.size:
            # no need for reduction
            lons, lats = np.meshgrid(x, y)
            return lons, lats, np.asarray(u), np.asarray(v)
        if nlon > x.size:
            nlon = x.size
        if nlat > y.size:
            nlat = y.size
        
        # create new arrays for a reduced number of vectors
        lon = np.zeros(nlon)
        lat = np.zeros(nlat)
        
        us = np.zeros((nlat, nlon))
        vs = np.zeros((nlat, nlon))
        
        # define intervals of a new grid
        dlon = int(floor(x.size / nlon))
        dlat = int(floor(y.size / nlat))
        dlon_half = int(floor(0.5*dlon))
        dlat_half = int(floor(0.5*dlat))
        
        # define a new grid
        for i in range(nlat):
            lat[i] = y[i*dlat + dlat_half]  
        for j in range(nlon):
            lon[j] = x[j*dlon + dlon_half]  
            
        lons,lats = np.meshgrid(lon, lat)

        # reduce a number of vectors, replacement by mean values
        for i in range(nlat):
            for j in range(nlon):
                u_temp = u[i*dlat:(i+1)*dlat, j*dlon:(j+1)*dlon]             
                v_temp = v[i*dlat:(i+1)*dlat, j*dlon:(j+1)*dlon]
                if np.all(np.isnan(u_temp)):
                    us[i,j] = np.nan
                    vs[i,j] = np.nan
                else:
                    us[i,j] = np.mean(u_temp)                           
                    vs[i,j] = np.mean(v_temp)

        return lons, lats, us, vs
    
    @classmethod
    def get_reference_arrow_data(cls, us, vs):
        """
        Gets parameters of a reference arrow.

        Parameters
        ----------
        us, vs : Numpy array
            2-dimensional arrays contain values of horizontal or vertical components of a variable.

        Returns
        -------                                  
        ref_arrow : int
            Value of a reference arrow.
        """
        
        # reference arrow = arrow with maximum length
        speed_max = np.nanmax(np.sqrt(us**2 + vs**2))

        # round and make a reference arrow multiple of 5            
        ref_arrow = int(np.round(speed_max/5.0)*5.0)
        
        return ref_arrow
    
    @classmethod
    @abstractmethod
    def make_colorbar(cls, colorset, valueset, tick_interval, variable_name, x_colorbar):
        """
        Creates a colorbar.

        Parameters
        ----------
        colorset : list of tuples
            Set of ordered unique colors. List containing tuples mapping a normalized value to a HEX or RGB. 
                 
        valueset : Numpy array
            Set of ordered unique double values.
                 
        tick_interval : float
            Value of intervals between ticks in a colorbar.
                      
        variable_name : str
            Name of a variable.
                      
        x_colorbar : float
            Position of a colorbar on axis X.

        Returns
        -------
        Plotly Scattergl or Bokeh ColorBar object
        """


"""
Basic plot routines usable to construct more advanced plots
"""
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from matplotlib.axes import Axes
import numpy as np
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from mpl_toolkits.axes_grid1.axes_size import AxesY, Fraction
from numba import jit
from multipledispatch import dispatch

from enstools.core import check_arguments
from enstools.misc import point_in_polygon
import xarray
import six
import math
import cartopy
import warnings


# names for coordinates
__lon_names = ["lon", "lons", "longitude", "longitudes", "clon", "rlon"]
__lat_names = ["lat", "lats", "latitude", "latitudes", "clat", "rlat"]

# cache for expensively calculated triangulations
__triangulation_cache = {}


def __is_global(lon, lat):
    """
    check if coordinates belong to a global dataset

    Parameters
    ----------
    lon : np.ndarray or xarray.DataArray
    lat : np.ndarray or xarray.DataArray

    Returns
    -------
    bool
    """
    if lon.max() - lon.min() > 350 and lat.max() - lat.min() > 170:
        return True
    return False


def __get_order_of_magnitude(value):
    """
    get the order of magnitude for a given value. This function may be used to create nice
    colorbar levels.

    Parameters
    ----------
    value : float

    Returns
    -------
    float

    Examples
    --------
    >>> __get_order_of_magnitude(3)
    5
    >>> __get_order_of_magnitude(7)
    10
    >>> __get_order_of_magnitude(0.7)
    1
    """
    x1 = [1, 2, 5]
    x2 = -10
    i = 0
    while True:
        x3 = x1[i] * 10**x2
        if x3 >= abs(value):
            break
        if i < 2:
            i += 1
        else:
            i = 0
            x2 = x2 + 1
    if value < 0:
        x3 = x3 * -1
    return x3


def get_nice_levels(variable, center_on_zero=False, nlevel=11, min_percentile=0.1, max_percentile=99.9):
    """
    Create levels for contour plots

    Parameters
    ----------
    variable : xarray.DataArray or np.ndarray
            the values for which the levels should be created

    nlevel : int
            number of levels

    max_percentile : float
            plots often look nicer if not the absolute maximum but is
            used, but a percentile.

    Returns
    -------
    np.ndarray
            level-array usable in matplotlib plotting methods
    """
    # use a percentile instead of the actual maximum
    minval, maxval = np.nanpercentile(variable, [min_percentile, max_percentile])
    perc10 = __get_order_of_magnitude(maxval - minval) / 10.0

    # adjust minimum, especially around zero
    minval_is_adjusted = False
    if minval > 0 > minval - perc10:
        minval = 0
        minval_is_adjusted = True
    elif minval < 0 < maxval and abs(minval) < abs(maxval) / 1000:
        minval = 0
        minval_is_adjusted = True
    elif center_on_zero:
        minval = 0
        minval_is_adjusted = True
    else:
        minval -= np.mod(minval, perc10)

    # adjust maximum
    maxval = maxval + perc10 - np.mod(maxval, perc10)
    step = __get_order_of_magnitude((maxval - minval) / nlevel)
    if not minval_is_adjusted and min_percentile > 0.0 and minval > variable.min():
        minval -= step
    # modify the maximum to ensure, that it is an integer multiple of the step
    range = maxval - minval
    if np.mod(range, step) != 0:
        maxval += step - np.mod(range, step)
    if center_on_zero:
        minval = -maxval
    return np.linspace(minval, maxval, nlevel)


@jit(nopython=True)
def __mask_triangles(triangles, proj_coords, boundary, mask):
    # coordinates are given as 2d-arrays
    x_coords_proj = proj_coords[:, 0]
    y_coords_proj = proj_coords[:, 1]

    # shrink the boundary a bit to be sure that no point is on the boundary
    boundary_x = np.empty(boundary.shape[0])
    boundary_y = np.empty(boundary.shape[0])
    offset = boundary.max() * 0.002
    for i in range(boundary.shape[0]):
        a = math.atan2(boundary[i, 1], boundary[i, 0])
        l = math.hypot(boundary[i, 0], boundary[i, 1]) - offset
        boundary_y[i] = math.sin(a) * l
        boundary_x[i] = math.cos(a) * l

    # mask all triangles with at least one point outside of the boundaries
    for icell in range(triangles.shape[0]):
        for iedge in range(3):
            if not point_in_polygon(boundary_x, boundary_y, x_coords_proj[triangles[icell, iedge]], y_coords_proj[triangles[icell, iedge]]):
                mask[icell] = True
                break


def __get_triangulation(projection, transformation, lon, lat, calculate_mask=True, cache=True):
    """
    calculate a triangulation for a given unstructured grid.

    Parameters
    ----------
    projection : cartopy.crs.Projection
            the target projection for the plot

    transformation : cartopy.crs.Projection
            the source projection of the coordinates.

    lon : xarray.DataArray or np.ndarray
            longitudes of the grid

    lat : xarray.DataArray or np.ndarray
            latitudes of the grid

    Returns
    -------
    mtri.Triangulation
            the calculation triangulation with points outside of the projection masked
    """
    # we need the coordinates as numpy arrays
    lon, lat = np.asarray(lon), np.asarray(lat)

    # is the result in cache?
    if cache:
        cache_key = (projection, transformation, sum(lon[:10]), sum(lat[:10]))
        if cache_key in __triangulation_cache:
            return __triangulation_cache[cache_key]

    # find the boundary of the projection as well as the grid coordinates in projection space
    boundary = np.asarray(projection.boundary)
    proj_coords = projection.transform_points(transformation, lon, lat)

    # calculate the triangulation
    tri = mtri.Triangulation(lon, lat)

    # mask triangles outside the projection
    if calculate_mask:
        mask = np.zeros(tri.triangles.shape[0], dtype=np.bool)
        __mask_triangles(tri.triangles, proj_coords, boundary, mask)
        tri.set_mask(mask)

    # cache the result
    if cache:
        __triangulation_cache[cache_key] = tri
    return tri


def get_coordinates_from_xarray(variable, lon_name=None, lat_name=None, create_mesh=True, only_spatial_dims=True, rad2deg=True):
    """
    get coordinate arrays from a xarray object

    Parameters
    ----------
    variable : xarray.DataArray
            xarray with coordinates

    lon_name : str
            name of the longitude coordinate

    lat_name : str
            name of the latitude coordinate

    rad2deg: bool
            convert radian to degrees. True by default.

    Returns
    -------
    tuple:
            lon, lat arrays
    """
    # check arguments
    if isinstance(lon_name, (np.ndarray, xarray.DataArray)) and isinstance(lat_name, (np.ndarray, xarray.DataArray)):
        return lon_name, lat_name
    if not isinstance(variable, xarray.DataArray):
        raise ValueError("named coordinates are only allowed for xarray variable!")

    # get the coordinates
    if lon_name is not None:
        lon = variable.coords[lon_name]
    else:
        lon = None
        for lon_name in __lon_names:
            if lon_name in variable.coords:
                lon = variable.coords[lon_name]
                break
        if lon is None:
            raise ValueError("not longitude coordinate found, checked names: %s" % ", ".join(__lon_names))
    if lat_name is not None:
        lat = variable.coords[lat_name]
    else:
        lat = None
        for lat_name in __lat_names:
            if lat_name in variable.coords:
                lat = variable.coords[lat_name]
                break
        if lat is None:
            raise ValueError("not longitude coordinate found, checked names: %s" % ", ".join(__lon_names))

    # check the units of the coordinates
    if rad2deg:
        if isinstance(lon, xarray.DataArray):
            if "units" in lon.attrs and lon.attrs["units"] == "radian":
                lon = np.rad2deg(lon)
        if isinstance(lat, xarray.DataArray):
            if "units" in lat.attrs and lat.attrs["units"] == "radian":
                lat = np.rad2deg(lat)

    # check the dimensions, only the dimensions of lon and lat are allowed
    if only_spatial_dims:
        coord_dim_names = []
        for dim_name in lon.dims:
            if dim_name not in coord_dim_names:
                coord_dim_names.append(dim_name)
        for dim_name in lat.dims:
            if dim_name not in coord_dim_names:
                coord_dim_names.append(dim_name)
        for dim_name in variable.dims:
            if dim_name not in coord_dim_names:
                raise ValueError("the variable has more dimensions (%s) then the horizontal coordinates (%s). Please call this function with a slice of the variable!" % (", ".join(variable.dims), ", ".join(coord_dim_names)))

    # do we need a mesh grid?
    if create_mesh and lon.ndim == 1 and variable.ndim == 2 and lon_name != "rlon" and lat_name != "rlat":
        lon, lat = np.meshgrid(lon, lat)
    return lon, lat


@check_arguments(dims={'variable': ('lat', 'lon')})
def contour(variable, lon=None, lat=None, **kwargs):
    """
    Create a plot from an xarray variable that includes coordinates.

    Examples
    --------

    >>> fig, ax = enstools.plot.contour(data["TOT_PREC"][0, :, :], coastlines="50m")    # doctest: +SKIP

    .. figure:: images/example_plot_icon_01.png

        24h ICON forecast for precipitation read from a grib2 file. Have a look at the script
        examples/example_plot_icon_01.py for more details.

    >>> fig, ax1 = enstools.plot.contour(data["PMSL"][0, :] / 100.0, gridlines=True, subplot_args=(121,))   # doctest: +SKIP
    >>> fig, ax2 = enstools.plot.contour(data["TOT_PREC"][0, :], figure=fig, subplot_args=(122,))           # doctest: +SKIP

    .. figure:: images/example_plot_icon_02.png

        24h ICON forecast for mean sea level pressure (left) and precipitation (right). The data was read from
        grib2 files on the native ICON grid and plotted without interpolation onto a regular grid. Have a look at the
        script examples/example_plot_icon_02.py

    Parameters
    ----------
    variable : xarray.DataArray
            the data to plot.

    lon : xarray.DataArray or np.ndarray or str
            longitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    lat : xarray.DataArray or np.ndarray or str
            latitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    Other optional keyword arguments:

    **kwargs
            *figure*: matplotlib.figure.Figure
                If provided, this figure instance will be used (and returned), otherwise a new
                figure will be created.

            *axes*: matplotlib.axes.Axes
                If provided, this axes instance will be used (e.g., of overplotting), otherwise a
                new axes object will be created.

            *subplot_args*: tuple
                Arguments passed on to add_subplot. These arguments are used only if no axes is provided.

            *subplot_kwargs*: dict
                Keyword arguments passed on to add_subplot. These arguments are used only if no axes is provided.

            *filled*: [*True* | *False*]
                If True a filled contour is plotted, which is the default

            *colorbar*: [*True* | *False* | "empty"]
                If True, a colorbar is created. Default=True. Use empty to reserve space for the colorbar
                without actually creating it. This is usefull for multipanel plots, where one panel has a colorbar and
                another not.

            *levels*: np.ndarray
                If provided, these levels are used, otherwise the levels are automatically selected.

            *levels_center_on_zero* : bool
                If true, automatically selected levels are centered around zero.

            *gridlines*: [*True* | *False*]
                If True, coordinate grid lines are drawn. Default=False

            *gridline_labes*: [*True* | *False*]
                Whether or not to label the grid lines. The default is not to label them.

            *coastlines*: [*True* | *False* | '110m' | '50m' | '10m']
                If True, coordinate grid lines are drawn. Default=True

            *coastlines_kwargs*: dict
                dictionary with arguments passed on to ax.coastlines()

            *borders*: [*True* | *False* | '110m' | '50m' | '10m']
                If True, coordinate grid lines are drawn. Default=False

            *projection*: [ *cartopy.crs.Projection*]
                If not None, the Projection object is used to create the plot

            *rotated_pole*: [*xarray.DataArray* | dict]
                Information about the rotated pole. This can either be the CF-standard rotated_pole variable from
                an input file, or alternatively a dictionary with the keys grid_north_pole_latitude and
                grid_north_pole_longitude.

    All other arguments are forwarded to the matplotlib contour or contourf function.

    Returns
    -------
    tuple
            (Figure, Axes) of the new plot is returned. The returned values may be reused in subsequent calls to
            plot functions.
    """
    # get the coordinates
    lon, lat = get_coordinates_from_xarray(variable, lon, lat)

    # is it a global dataset?
    transformation = ccrs.PlateCarree()
    is_global = __is_global(lon, lat)
    projection = kwargs.get("projection", None)
    if projection is None:
        if is_global:
            projection = ccrs.Mollweide()
        else:
            projection = ccrs.PlateCarree()

    # is it a rotated pole array
    if kwargs.get("rotated_pole", None) is not None:
        rotated_pole = kwargs.get("rotated_pole")
        if isinstance(rotated_pole, xarray.DataArray):
            grid_north_pole_latitude = rotated_pole.grid_north_pole_latitude
            grid_north_pole_longitude = rotated_pole.grid_north_pole_longitude
        elif isinstance(rotated_pole, dict):
            grid_north_pole_latitude = rotated_pole["grid_north_pole_latitude"]
            grid_north_pole_longitude = rotated_pole["grid_north_pole_longitude"]
        else:
            raise ValueError("unsupported type of rotated_pole argument!")

        # create a projection if no projection was specified
        if not "projection" in kwargs:
            projection = ccrs.RotatedPole(grid_north_pole_longitude, grid_north_pole_latitude)

        # create a transformation if the coordinates are 1d
        if lon.ndim == 1:
            transformation = projection

    # create the plot
    if kwargs.get("figure", None) is None:
        fig = plt.figure()
    else:
        fig = kwargs.pop("figure")
    if kwargs.get("axes", None) is None:
        subplot_args = kwargs.pop("subplot_args", (111,))
        subplot_kwargs = kwargs.pop("subplot_kwargs", {})
        ax = fig.add_subplot(*subplot_args, projection=projection, **subplot_kwargs)
    else:
        ax = kwargs.pop("axes")

    # construct arguments for the matplotlib contour
    contour_args = {"transform": transformation,
                    "extend": "both"}
    if kwargs.get("levels", None) is None:
        contour_args["levels"] = get_nice_levels(variable, center_on_zero=kwargs.get("levels_center_on_zero", False))
    else:
        contour_args["levels"] = kwargs["levels"]
    if kwargs.get("cmap", None) is not None:
        contour_args["cmap"] = kwargs["cmap"]
    elif kwargs.get("colors", None) is None:
        contour_args["cmap"] = "CMRmap_r"

    # copy all arguments not specific to this function to the contour arguments
    for arg, value in six.iteritems(kwargs):
        if arg not in ["coastlines", "coastlines_kwargs", "borders", "borders_kwargs", "filled", "colorbar", "gridlines", "gridline_labes", "projection", "levels_center_on_zero", "rotated_pole", "cmap"]:
            contour_args[arg] = value

    # decide on the plot type based on variable and coordinate dimension
    # FIXME: suppress a specific warning about MaskedArrays
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', lineno=6385)  # suppress a specific warning about MaskedArrays
        warnings.filterwarnings('ignore', lineno=6442)  # suppress a specific warning about MaskedArrays

        if variable.ndim == 1 and lon.ndim == 1 and lat.ndim == 1:
            # calculate the triangulation
            tri = __get_triangulation(projection, transformation, lon, lat)
            # create the plot
            if kwargs.get("filled", True):
                contour = ax.tricontourf(tri, variable, **contour_args)
            else:
                contour = ax.tricontour(tri, variable, **contour_args)

        else:
            # create a contour plot
            # transpose if necessary, lat is expected to be the first dimension of the variable
            if kwargs.get("filled", True):
                contour = ax.contourf(lon, lat, variable, **contour_args)
            else:
                contour = ax.contour(lon, lat, variable, **contour_args)

    # add coastlines
    # TODO: ensure that coastlines or borders are not added multiple times
    resolution_coastlines = kwargs.get("coastlines", True)
    if resolution_coastlines is not False:
        coastlines_kwargs = kwargs.get("coastlines_kwargs", dict())
        if isinstance(resolution_coastlines, str):
            ax.coastlines(resolution_coastlines, **coastlines_kwargs)
        else:
            ax.coastlines(**coastlines_kwargs)

    # add borders
    resolution_borders = kwargs.get("borders", False)
    if not isinstance(resolution_borders, str) and isinstance(resolution_coastlines, str):
        resolution_borders = resolution_coastlines
    if resolution_borders is not False:
        if isinstance(resolution_borders, str):
            ax.add_feature(cartopy.feature.NaturalEarthFeature(
                "cultural", "admin_0_boundary_lines_land",
                resolution_borders, edgecolor='black', facecolor='none', linestyle=":"))
        else:
            ax.add_feature(cartopy.feature.BORDERS, linestyle=":")

    # add gridlines
    if kwargs.get("gridlines", False) is True:
        ax.gridlines(color="black", linestyle="dotted", draw_labels=kwargs.get("gridline_labels", False))

    # add a colorbar
    if kwargs.get("colorbar", True) is True or kwargs.get("colorbar", True) == "empty":
        divider = make_axes_locatable(ax)
        width = AxesY(ax, aspect=0.07)
        pad = Fraction(1.0, width)
        cb_ax = divider.append_axes("right", size=width, pad=pad, axes_class=Axes, frameon=False)
        if kwargs.get("colorbar", True) is True:
            fig.colorbar(contour, cax=cb_ax)
        else:
            cb_ax.tick_params(bottom="off", left="off", top="off", right="off", which="both", labelbottom="off", labelleft="off")

    if is_global:
        ax.set_global()

    return fig, ax



from enstools.core import check_arguments
from enstools.misc import distance
import numpy as np
from scipy.fft import dctn, idctn
from numba import njit, objmode


@check_arguments(dims={'variable': ('lat', 'lon')})
def dct_2d_regional(variable, lon, lat, high_cutoff=0.0, low_cutoff=0.0):
    """
    Filter an input array on a regular lat-lon grid using a discrete cosine transformation

    Parameters
    ----------
    variable: xarray.DataArray
        2d variable on a regular lat-lon grid

    lon: xarray.DataArray
        longitude of input data, 1D.

    lat: xarray.DataArray
        latitude of input data, 1D.

    low_cutoff: float
        wave length for low-pass filter in km. Shorter wave lengths are removed. The default
        value of zero means, that no filter is applied.

    high_cutoff: float
        wave length for high-pass filter in km. Longer wave lengths are removed. The default
        value of zero means, that no filter is applied.

    Returns
    -------
    filtered: xarray.DataArray
        filtered version of the input
    """

    # forward pass of DCT on the input data
    coefficients = dctn(variable.values, type=2)

    # apply the actual filter
    __dct_2d_regional(coefficients, lon.values, lat.values, high_cutoff, low_cutoff)

    # perform the inverse transformation
    filtered = idctn(coefficients, type=2)

    # pack result into xarray object
    result = variable.copy(data=filtered)
    return result


@njit()
def __dct_2d_regional(transformed_array, lon, lat, high_cutoff=0.0, low_cutoff=0.0):

    # low-pass filter
    if low_cutoff > 0:
        # calculate the wave number corresponding to the cut-off wave length
        wn_lon, wn_lat = get_wavenumber_from_length(low_cutoff, lon, lat)

        # use an elliptic mask to filter the coefficients (compare Denis et al. 2002)
        # set all the coefficients outside the ellipse to 0
        for i in range(len(lat)):
            for j in range(len(lon)):
                # No idea why the Python and Fortran wn_pos different in the factor 2 in the end here:
                wn_pos = np.sqrt(float(j)**2 / (wn_lon-1)**2 + float(i)**2 / (wn_lat-1)**2)  
                if wn_pos > 1.:
                    transformed_array[i,j] = 0.

    # high-pass filter
    if high_cutoff > 0:
        # calculate the wave number corresponding to the cut-off wave length
        wn_lon, wn_lat = get_wavenumber_from_length(high_cutoff, lon, lat)

        # use an elliptic mask to filter the coefficients (compare Denis et al. 2002)
        # set all the coefficients inside the ellipse to 0
        for i in range(len(lat)):
            for j in range(len(lon)):
                # No idea why the Python and Fortran wn_pos different in the factor 2 in the end here:
                wn_pos = np.sqrt(float(j)**2 / (wn_lon-1)**2 + float(i)**2 / (wn_lat-1)**2)
                if wn_pos <= 1.:
                    transformed_array[i,j] = 0.
    return transformed_array


@njit()
def get_wavenumber_from_length(cutoff_length, coord_lon, coord_lat):
    """
    Find the wave number that corresponds to a given length in lon- and lat-direction.
    """

    # calculate the lat-extent of the array along an axis that is centered in the domain
    length = 0.
    x = int(len(coord_lon) / 2) # maybe +1 
    for i in range(len(coord_lat)-1):
        length += distance(coord_lat[i], coord_lat[i+1], coord_lon[x], coord_lon[x], input_in_radian=False)
    wave_number_lat = np.rint(length / 1000.0 / cutoff_length * 2) + 1

    # calculate the lon-extent of the array along an axis that is centered in the domain
    length = 0.
    y = int(len(coord_lat) / 2) # maybe +1
    for i in range(len(coord_lon)-1):
        length += distance(coord_lat[y], coord_lat[y], coord_lon[i], coord_lon[i+1], input_in_radian=False)
    wave_number_lon = np.rint(length / 1000.0 / cutoff_length * 2) + 1

    # limit the wave numbers
    if wave_number_lon < 2:
        wave_number_lon = 2
    if wave_number_lat < 2:
        wave_number_lat = 2
    return wave_number_lon, wave_number_lat
